"""
Python 2 & 3 Compatibility Functions

"""


def ascii(value):
    """Return the string of value

    :param mixed value: The value to return
    :rtype: str

    """
    return '{0}'.format(value)
